# from k2eg.dml import k2eg
import os
import click
import logging
import k2eg.__version__ as __version__
import k2eg.cli.get as get
import k2eg.cli.put as put
import k2eg.cli.monitor as monitor
import k2eg
from click_loglevel import LogLevel
from click_repl import repl
from click_repl import register_repl

k2eg_dml_instance: k2eg.dml = None
initilized: bool = False
in_shell: bool = False

@click.group(chain=True, invoke_without_command=True)
@click.pass_context
@click.version_option(__version__.__version__, prog_name="K2EG demo cli")
@click.option('-e', '--environment',  
              required=True, 
              type=str, 
              envvar='K2EG_CLI_DEFAULT_ENVIRONMENT')
@click.option('-a', '--app-name',  
              required=True, 
              type=str, 
              envvar='K2EG_CLI_DEFAULT_APP_NAME')
@click.option(
    "-l", "--log-level",
    type=LogLevel(extra=["VERBOSE", "NOTICE"]),
    default=logging.INFO,
)
def cli(ctx, environment, app_name, log_level):
    global k2eg_dml_instance
    global initilized
    global in_shell
    if not initilized:
        ctx.obj = {}
        logging.basicConfig(
            format="[%(levelname)-8s] %(message)s",
            level=log_level,
        )
        if os.environ.get('K2EG_PYTHON_CONFIGURATION_PATH_FOLDER') is None:
            config_path = "default embedded"
        else:
            config_path = os.environ.get('K2EG_PYTHON_CONFIGURATION_PATH_FOLDER')
        logging.debug("Use configuration '{}'".format(config_path))
        logging.debug("Use environment '{}'".format(environment))
        logging.debug("Use application name '{}'".format(app_name))
        # allocate k2eg in 
        logging.debug("K2g initilizing")
        ctx.obj = k2eg_dml_instance = k2eg.dml(environment, app_name)
        initilized = True
    
    if ctx.invoked_subcommand is None:
        in_shell = True
        repl(ctx)
        print('Exing shell')
        in_shell = False
        
    pass

@cli.result_callback()
def process_pipeline(processors, environment, app_name, log_level):
    if k2eg_dml_instance is not None and in_shell is False:
        logging.debug("Deinit dml")
        k2eg_dml_instance.close()
        logging.debug("Closed dml")

register_repl(cli)  # Register the REPL command
cli.add_command(get.get)
cli.add_command(put.put)
cli.add_command(monitor.monitor)
if __name__ == "__main__":
    cli()


from dataclasses import dataclass
import datetime
import json
import os
import logging
import re
import threading
import configparser
import time
from confluent_kafka import Consumer, TopicPartition, Producer, OFFSET_END, OFFSET_BEGINNING
from confluent_kafka import KafkaError, KafkaException

@dataclass
class SnapshotProperties:
    """
    Class to manage the snapshot properties
    """
    # The name of the snapshot
    snapshot_name:str = None
    # The time window for the snapshot
    time_window: int = 1000
    # The delay between snapshots
    repeat_delay: int = 0
    # Define if a snapshot need to be automatically or manually triggered
    triggered: bool = False
    # The list of PV URIs to snapshot
    pv_uri_list:list[str] = None
    
    def validate(self):
        """
        Validate the snapshot properties.

        Raises:
            ValueError: If any property is invalid.
        """
        if not self.snapshot_name or not isinstance(self.snapshot_name, str):
            raise ValueError("snapshot_name must be a non-empty string.")
        if not re.match(r'^[A-Za-z0-9_\-]+$', self.snapshot_name):
            raise ValueError(
                f"Invalid snapshot name '{self.snapshot_name}'. Only alphanumeric characters, dashes, and underscores are allowed."
            )
        if not self.pv_uri_list or not isinstance(self.pv_uri_list, list):
            raise ValueError("pv_uri_list must be a non-empty list of PV URIs.")
        if not all(isinstance(pv, str) for pv in self.pv_uri_list):
            raise ValueError("All PV URIs in pv_uri_list must be strings.")
        if not isinstance(self.time_window, int) or self.time_window < 0:
            raise ValueError("time_window must be a non-negative integer.")
        if not isinstance(self.repeat_delay, int) or self.repeat_delay < 0:
            raise ValueError("repeat_delay must be a non-negative integer.")
        if not isinstance(self.triggered, bool):
            raise ValueError("triggered must be a boolean value.")

class TopicUnknown(Exception):
    """Exception raised when the timeout is expired on operation"""
    def __init__(self, message):            
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

class TopicChecker:
    def __init__(self):
        self.__topics_to_check = []
        self.__check_timeout = None
        self.__mutex = threading.Lock()

    def __check_for_topic(self, topic_name, consumer:Consumer):
        cluster_metadata = consumer.list_topics(topic_name, 1.0)
        return cluster_metadata.topics[topic_name].error \
                != KafkaError.UNKNOWN_TOPIC_OR_PART

    def add_topic(self, topic_name):
        with self.__mutex:
            logging.debug(f"Add topic {topic_name} to checker")
            self.__topics_to_check.append(topic_name)

    def remove_topic(self, topic_name):
        with self.__mutex:
            if topic_name in self.__topics_to_check:
                logging.debug(f"remove topic {topic_name} to checker")
                self.__topics_to_check.remove(topic_name)
            else:
                logging.debug(f"topic {topic_name} not in checker")

    def update_metadata(self, consumer: Consumer) -> bool:
        to_check =  self.__check_timeout is None or \
                    (self.__check_timeout is not None and \
                    datetime.datetime.now() > self.__check_timeout)
        if to_check:
            with self.__mutex:
                for t in self.__topics_to_check:
                    found = self.__check_for_topic(t, consumer)
                    logging.debug(f"Check for topic {t} => found:{found}")
                    # low, high = consumer.get_watermark_offsets(t)
                    # logging.debug(f'Found max and min [{high},{low}] index for topic: {t}')
                    if found:
                        logging.debug(f"Remove topic {t} from checker")
                        self.__topics_to_check.remove(t)

                if(len(self.__topics_to_check)>0):
                    self.__check_timeout = datetime.datetime.now() \
                        + datetime.timedelta(seconds=3)

class Broker:
    """
    Represent the abstraction of the protocol 
    which the k2eg client talk with the k2eg 
    gateway

    ...

    Attributes
    ----------

    Methods
    -------
    """
    def __init__(
        self, 
        environment_id: str,
        app_name:str,
        group_name: str = None,
        app_instance_unique_id:str = "1",
        startup_tout: int = 10):
        """
        Parameters
        ----------
        environment_id : str
            The name of the environment to use, this wil be the danem of the
            ini file where to load the configuration. (<environment_id>.ini)
        group_name : str
            Is the group name to distribuite the data from different same 
            client instance
        """
        if group_name is None:
            group_name = app_name
        enviroment_set: str = 'DEFAULT'
        self.__initialized=False
        # Get the current directory of the script
        current_configuration_dir = os.getenv(
            'K2EG_PYTHON_CONFIGURATION_PATH_FOLDER', 
            os.path.dirname(os.path.realpath(__file__))+"/environment"
        ) 
        enable_kafka_debug = os.getenv(
            'K2EG_PYTHON_ENABLE_KAFKA_DEBUG_LOG', 
            'false'
        ).lower in ("yes", "true", "t", "1")

        self.__enviroment_set = enviroment_set
        # Create a new ConfigParser object
        self.__config = configparser.ConfigParser()

        # Read the configuration file
        self.__config.read(
            os.path.join(
            current_configuration_dir, 
            f'{environment_id}.ini')
            )

        self.__check_config(app_name)

        config_consumer = {
            'bootstrap.servers': self.__config.get(
                self.__enviroment_set, 'kafka_broker_url'
                ), 
            'group.id': group_name,
            'group.instance.id': app_name+'_'+app_instance_unique_id,
            'auto.offset.reset': 'latest',
            'enable.auto.commit': 'true',
            'topic.metadata.refresh.interval.ms': '60000',
        }
        if enable_kafka_debug is True:
            config_consumer['debug'] = 'consumer,fetch'

        self.__consumer = Consumer(config_consumer)
        config_producer = {
            'bootstrap.servers': self.__config.get(
                self.__enviroment_set, 'kafka_broker_url'
                ),
            #'debug': 'consumer,cgrp,topic,fetch',
        }
        self.__producer = Producer(config_producer)
        self.__reply_topic = app_name + '-reply'
        self.__reply_partition_assigned = threading.Event()
        self.__subribed_topics = [self.__reply_topic]

        # start consuming from reply topic
        self.__consumer.subscribe(self.__subribed_topics, on_assign=self.__on_assign)
        self.__reply_topic_joined = False
        self.__topic_checker = TopicChecker()
        # wait for consumer join the partition
        self.wait_for_reply_available(startup_tout)
        

    def __manage_old_state(self):
        self.__consumer.poll(1.0)
        assigned_topics = self.__consumer.assignment()
        for t in assigned_topics:
            logging.debug(f'latest assigned topic: {t}')
        self.__consumer.unsubscribe()

    # point always to the end of the topic
    def __on_assign(self, consumer, partitions):
        for p in partitions:
            logging.debug(f"Joined topic {p.topic} with partition {p.partition}")
            if p.topic == self.__reply_topic:
                self.__reply_topic_joined = True
                self.__reply_partition_assigned.set()
                #if p.offset==-1 or p.offset==-1001:
                logging.debug(f'Force to reading from the end for topic: {p.topic}')
                #p.offset = OFFSET_END
            else:
                try:
                    low, high = consumer.get_watermark_offsets(p)
                    logging.debug(f'Found max and min [{high},{low}] index for topic: {p.topic}')
                    # in this case we have to go one index behing to start reading from the
                    # last element in the queue
                    if high >= 1:
                        # new_offset = high
                        logging.debug(f"offset for topic '{p.topic}' is '{high}'")
                        p.offset = OFFSET_END
                    elif high < 0:
                        p.offset = OFFSET_END
                    else:
                        p.offset = OFFSET_BEGINNING
                except Exception as e:
                    logging.debug(f'got exception on metadata refresh: {e}')
                    p.offset = OFFSET_END       
        consumer.assign(partitions)


    def __check_config(self, app_name):
        if not self.__config.has_option(self.__enviroment_set, 'kafka_broker_url'):
            raise ValueError('[kafka_broker_url] Kafka broker url is mandatory')
        if not self.__config.has_option(self.__enviroment_set, 'k2eg_cmd_topic'):
            raise ValueError("[k2eg_cmd_topic] The topic "
                             "for send command to k2eg is mandatory")

    def get_reply_topic(self):
        return self.__reply_topic

    def wait_for_reply_available(self, timeout):
        """ Wait untile the consumer has joined the reply topic
        """
        start_time = time.time()
        end_time = start_time + timeout
        while self.__reply_partition_assigned.wait(1) is False:
            if time.time() > end_time:
                raise TimeoutError("Function timed out")
            logging.debug("waiting for reply topic to join")
            self.__consumer.poll(1)

    def reset_reply_topic_to_ts(self, timestamp):
        self.reset_topic_offset_in_time(self.__reply_topic, timestamp)

    def reset_topic_offset_in_time(self, topic, timestamp):
        """ Set the topic offset to the specific timestamp
        """
        partitions = self.__consumer.list_topics(topic).topics[topic].partitions.keys()

        # Create TopicPartition objects for each partition, with the specific timestamp
        topic_partitions = [
            TopicPartition(topic, p, int(timestamp * 1000)) for p in partitions
        ]

        # Get the offsets for the specific timestamps
        try:
            offsets_for_times = self.__consumer.offsets_for_times(topic_partitions)
            # Set the starting offset of the consumer to the returned offsets
            for tp in offsets_for_times:
                if tp.offset != -1 or tp.offset != -1001:  # If an offset was found
                    self.__consumer.seek(tp)
                else:
                    tp.offset = OFFSET_BEGINNING
        except KafkaException as e:
            logging.error(e)
       
    def get_next_message(self, timeout = 0.300):
        message = self.__consumer.poll(timeout=timeout)
        # give a chanche to update metadata ofr pending topics
        self.__topic_checker.update_metadata(self.__consumer)
        if message is None:
            return None
        # tp = TopicPartition(message.topic(), message.partition(), message.offset() + 1)
        # # Asynchronous commit
        # self.__consumer.commit(offsets=[tp], asynchronous=True)
        if message.error():
            if message.error().code() == KafkaError.UNKNOWN_TOPIC_OR_PART:
                logging.info(f"Topic {message.topic()} not found, add it to checker")
                self.__topic_checker.add_topic(message.topic())
        return message
    
    def commit_current_fetched_message(self):
        self.__consumer.commit()

    def add_topic(self, new_topic):
        if new_topic == self.__reply_topic:
            raise ValueError(
                f'The topic name {self.__reply_topic} cannot be used'
                )
        if new_topic not in self.__subribed_topics:
            logging.debug(f'Start consuming from topic: {new_topic}')
            self.__subribed_topics.append(new_topic)
            self.__consumer.subscribe(
                self.__subribed_topics, on_assign=self.__on_assign)
        else:
            logging.debug(f'Topic {new_topic} is already consuming')

    def remove_topic(self, topic_to_remove):
        if topic_to_remove == self.__reply_topic:
            raise ValueError(
                f'The topic name {self.__reply_topic} cannot be used'
                )
        self.__topic_checker.remove_topic(topic_to_remove)
        if topic_to_remove in self.__subribed_topics:
            logging.debug(f'Topic {topic_to_remove} will be removed')
            self.__subribed_topics.remove(topic_to_remove)
            self.__consumer.subscribe(
                self.__subribed_topics, on_assign=self.__on_assign
                )
        else:
            logging.debug(f'Topic {topic_to_remove} is not present')
  
    def send_command(self, message: str):
        broker_cmd_in_topic = self.__config.get(self.__enviroment_set, 'k2eg_cmd_topic')
        self.__producer.produce(
            broker_cmd_in_topic,
            message
        )
        self.__producer.flush()

    def send_get_command(self, pv_uri, reply_id):
        json_msg = {
            "command": "get",
            "serialization": "msgpack",
            "pv_name": pv_uri,
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id
        }
        self.send_command(json.dumps(json_msg))

    def send_start_monitor_command(self, pv_uri, pv_reply_topic, reply_id):
        # send command
        json_msg = {
            "command": "monitor",
            "serialization": "msgpack",
            "pv_name": pv_uri,
            "activate": True,
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id,
            "monitor_dest_topic": pv_reply_topic
        }
        self.send_command(json.dumps(json_msg))    
    
    def send_start_monitor_command_many(self, pv_uri_list, reply_id):
        # send command
        json_msg = {
            "command": "multi-monitor",
            "serialization": "msgpack",
            "activate": True,
            "pv_name_list": pv_uri_list,
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id
        }
        self.send_command(json.dumps(json_msg))  

    def send_put_command(self, pv_uri: str, value: any, reply_id: str):
        json_msg = {
            "command": "put",
            "pv_name": pv_uri,
            "value": str(value),
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id,
            "serialization": "msgpack"
        }
        self.send_command(json.dumps(json_msg))   

    def send_snapshot_command(self, pv_uri_list:list[str], reply_id:str):
        json_msg = {
            "command": "snapshot",
            "serialization": "msgpack",
            "pv_name_list": pv_uri_list,
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id
        }
        self.send_command(json.dumps(json_msg))  

    def send_repeating_snapshot_command(self, properties:SnapshotProperties, reply_id:str):
        #validate properties
        properties.validate()
        json_msg = {
            "command": "repeating_snapshot",
            "serialization": "msgpack",
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id,
            "snapshot_name": properties.snapshot_name,
            "pv_name_list": properties.pv_uri_list,
            "repeat_delay_msec": properties.repeat_delay,
            "time_window_msec": properties.time_window,
            "triggered": properties.triggered
        }
        self.send_command(json.dumps(json_msg))  
    
    def send_repeating_snapshot_trigger_command(self, snapshot_name:str, reply_id:str):
        #validate properties
        json_msg = {
            "command": "repeating_snapshot_trigger",
            "serialization": "msgpack",
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id,
            "snapshot_name": snapshot_name,
        }
        self.send_command(json.dumps(json_msg)) 
    
    def send_repeating_snapshot_stop_command(self, snapshot_name:str, reply_id:str):
        #validate properties
        json_msg = {
            "command": "repeating_snapshot_stop",
            "serialization": "msgpack",
            "reply_topic": self.__reply_topic,
            "reply_id": reply_id,
            "snapshot_name": snapshot_name,
        }
        self.send_command(json.dumps(json_msg)) 

    def initialized(self):
        if not self.self.__initialized:
            raise 

    def close(self):
        self.__producer.flush()
        self.__consumer.close()


import re
import uuid
import msgpack
import logging
import threading
import datetime
from enum import Enum
from time import sleep
from readerwriterlock import rwlock
from confluent_kafka import KafkaError
from k2eg.broker import Broker, SnapshotProperties
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import Callable, List, Dict, Any

_protocol_regex = r"^(pva?|ca)://((?:[A-Za-z0-9-_:]+(?:\.[A-Za-z0-9-_]+)*))$"

def _filter_pv_uri(uri: str):
    match = re.match(_protocol_regex, uri)
    if match:
        return match.group(1), match.group(2)
    else:
        return None, None

class OperationTimeout(Exception):
    """Exception raised when the timeout is expired on operation"""
    def __init__(self, message):            
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

class OperationError(Exception):
    """Exception raised when the timeout is expired on operation"""
    def __init__(self, error, message):            
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        self.error = error

class SnapshotState(Enum):
    INITIALIZED = 0
    HEADER_RECEVED = 1
    DATA_ACQUIRING = 2
    TAIL_RECEIVED = 3

@dataclass
class Snapshot:
    handler: Callable[[str, Dict[str, Any]], None]
    properties: SnapshotProperties = None
    publishing_topic: str = None
    state: SnapshotState = SnapshotState.INITIALIZED 
    timestamp: datetime.datetime = None
    interation: int = 0
    pv_list: List[str] = field(default_factory=list)
    results: List[Dict[str, Any]] = field(default_factory=list)

class dml:
    """K2EG client"""
    def __init__(
            self, 
            environment_id: str, 
            app_name: str,
            group_name: str = None):
        if app_name is None:
            raise ValueError(
                "The app name is mandatory"
            )
        self.__broker = None
        self.__thread = None
        self.__broker = Broker(environment_id, app_name, group_name)
        self.__lock = rwlock.RWLockFairD()
        self.__reply_partition_assigned = threading.Event()
        
        #reset to listen form now
        self.__consume_data = True
        self.__thread = threading.Thread(
            target=self.__consumer_handler
        )
        self.__thread.start()
        self.__monitor_pv_handler = {}
        self.reply_wait_condition = threading.Condition()
        self.reply_ready_event = threading.Event()
        self.reply_message = {}
        #contain a vector for each reply id where snapshot are stored
        self.reply_snapsthot_message = {}
        self.reply_recurring_snapsthot_message = {}

    def __del__(self):
        # Perform cleanup operations when the instance is deleted
        self.close()

    def __from_json(self, j_msg):
        print('__from_json')

    def __from_msgpack(self, m_msg):
        msg_id = None
        converted_msg = None
        decodec_msg = msgpack.loads(m_msg)
        if not isinstance(decodec_msg, dict):
            return msg_id, converted_msg
        
        if 'reply_id' in decodec_msg:
            msg_id = decodec_msg['reply_id']
            converted_msg = decodec_msg
        elif 'snapshot_name' in decodec_msg:
            msg_id = decodec_msg['snapshot_name']
            converted_msg = decodec_msg
        else:
            msg_id = list(decodec_msg.keys())[0]
            converted_msg = decodec_msg
        
        # Add message-size key with the received msgpack size
        converted_msg['message-size'] = len(m_msg)
        return msg_id, converted_msg

    def __from_msgpack_compack(self, mc_msg):
        print('__from_msgpack_compack')

    def __decode_message(self, msg):
        """ Decode single message
        """
        msg_id = None
        converted_msg = None
        headers = msg.headers()
        if headers is None:
            logging.error("Message without header received")
            return msg_id, converted_msg
        
        for key, value in headers:
            if key == 'k2eg-ser-type':
                st = value.decode('utf-8').lower()
                if st == "json":
                    msg_id, converted_msg = self.__from_json(
                        msg.value()
                        )
                elif st == "msgpack":
                    msg_id, converted_msg = self.__from_msgpack(
                        msg.value()
                        )
                elif st == "msgpack-compact":
                    msg_id, converted_msg = self.__from_msgpack_compack(
                        msg.value()
                        )
                break   
        return msg_id, converted_msg

    def process_event(self, topic_name, msg_id, decoded_message):
        logging.debug(f"received event on topic {topic_name}")
        self.__monitor_pv_handler[msg_id](msg_id, decoded_message)
    
    def __consumer_handler(self):
        """ Consume message form kafka consumer
        after the message has been consumed the header 'k2eg-ser-type' is checked 
        for find the serialization:
            json, 
            msgpack, 
            msgpack-compact
        """
        with  ThreadPoolExecutor(max_workers=10) as executor:
            while self.__consume_data:
                message = self.__broker.get_next_message()
                if message is None: 
                    continue
                if message.error():
                    if message.error().code() == KafkaError._PARTITION_EOF:
                        # End of partition event
                        logging.error(
                            f"{message.topic()} [{message.partition()}]reached "+
                            f"end at offset {message.offset()}"
                        )
                    else:
                        continue
                else:
                    was_a_reply = False
                    from_topic = message.topic()
                    #msg_id could be a reply id or pv name
                    msg_id, decoded_message = self.__decode_message(message)
                    if msg_id is None or decoded_message is None:
                        continue
                    with self.reply_wait_condition:
                        was_a_reply = msg_id in self.reply_message
                        if was_a_reply is True:
                            # print(f"message received from topic: {message.topic()} offset: {message.offset()}")
                            logging.debug(f"received reply on topic {message.topic()}")
                            self.reply_message[msg_id] = decoded_message
                            self.reply_wait_condition.notify_all()
                        elif msg_id in self.__monitor_pv_handler:
                                executor.submit(
                                    self.process_event,
                                    message.topic(),
                                    msg_id,
                                    decoded_message[msg_id]
                                )
                        elif msg_id in self.reply_snapsthot_message:
                            # if the message is not a reply and not a monitor
                            # it should be a snapshot
                            snapshot = self.reply_snapsthot_message[msg_id]
                            # check if the message is a snapshot completion error == 1
                            if decoded_message.get('error', 0) == 0:
                                logging.debug(f"Added message to snapshot {msg_id}]")  
                                # the message contains a snapshot value
                                snapshot.results.append(decoded_message)
                            else:
                                logging.debug(f"Snapshot {msg_id} compelted with error {decoded_message.get('error', 0)}")
                                # we got the completion message so             
                                # remove the snapshot from the list
                                del self.reply_snapsthot_message[msg_id]
                                # and call async handler in another thread
                                executor.submit(
                                    snapshot.handler,
                                    msg_id,
                                    snapshot.results
                                )
                        elif from_topic in self.reply_recurring_snapsthot_message:
                            # iit should be a recurring snapshot
                            snapshot = self.reply_recurring_snapsthot_message[from_topic]
                            message_type = decoded_message.get('message_type', None)
                            if message_type is None:
                                continue
                            if message_type == 0 and (snapshot.state == SnapshotState.INITIALIZED or snapshot.state == SnapshotState.TAIL_RECEIVED):
                                snapshot.state = SnapshotState.HEADER_RECEVED
                                #get the timestsamp and iteration
                                snapshot.timestamp = decoded_message.get('timestamp', None)
                                snapshot.interation = decoded_message.get('iter_index', 0)
                                logging.debug(f"recurring snapshot {from_topic} header received [ state {snapshot.state}] messages {len(snapshot.results)}")
                            elif message_type == 1 and (snapshot.state  == SnapshotState.HEADER_RECEVED or snapshot.state  == SnapshotState.DATA_ACQUIRING):
                                ## we are acquireing data for the snapshtot
                                snapshot.state  == SnapshotState.DATA_ACQUIRING
                                snapshot.results.append(decoded_message)
                                logging.debug(f"recurring snapshot {from_topic} data received [ state {snapshot.state}] messages {len(snapshot.results)}")
                            elif message_type == 2 and (snapshot.state == SnapshotState.HEADER_RECEVED or snapshot.state == SnapshotState.DATA_ACQUIRING):
                                # we got the completion message on snapshot that we are managing         
                                # renew the snapshot
                                snapshot.state = SnapshotState.TAIL_RECEIVED
                                logging.debug(f"recurring snapshot {from_topic} tail received [ state {snapshot.state}] messages {len(snapshot.results)}")  
                                # Prepare a dictionary with only PV values, inter_index, and timestamp
                                handler_data = {}
                                # Add 'iter_index' and 'timestamp' if present and not already set
                                handler_data["iteration"] = snapshot.interation
                                handler_data["timestamp"] =  snapshot.timestamp
                                for result in snapshot.results:
                                    # Extract PV values comparing with the list of PVs
                                    for pv_name in snapshot.pv_list:
                                        if pv_name in result:
                                            handler_data[pv_name] = result[pv_name]

                                # remove all received messages
                                snapshot.results = []

                                # and call async handler in another thread
                                executor.submit(
                                    snapshot.handler,
                                    msg_id,
                                    handler_data
                                )
                                    
                            else:
                                #log the error
                                logging.error(
                                    f"Error during snapshot {msg_id} with message type {message_type} and state {snapshot.state}"
                                )



    def parse_pv_url(self, pv_url):
        protocol, pv_name = _filter_pv_uri(pv_url)
        if protocol is None  or pv_name is None:
            raise ValueError(
                "The url is not well formed"
            )
        return protocol, pv_name

    def __check_pv_name(self, pv_url):
        pass

    def _check_pv_list(self, pv_uri_list: list[str]):
        for pv_uri in pv_uri_list:
            protocol, pv_name = self.parse_pv_url(pv_uri)
            if protocol.lower() not in ("pva", "ca"):
                raise ValueError("The protocol need to be one of 'pva'  'ca'")
            
    def __normalize_pv_name(self, pv_name):
        return pv_name.replace(":", "_")

    def _validate_snapshot_name(self, snapshot_name: str) -> None:
        """
        Validate the snapshot name. Only alphanumeric characters, dashes, and underscores are allowed.

        Args:
            snapshot_name (str): The snapshot name to validate.

        Raises:
            ValueError: If the snapshot name contains invalid characters.
        """
        if not re.match(r'^[A-Za-z0-9_\-]+$', snapshot_name):
            raise ValueError(
                f"Invalid snapshot name '{snapshot_name}'. Only alphanumeric characters, dashes, and underscores are allowed."
            )

    def __wait_for_reply(self, new_reply_id, timeout) -> tuple[int, any]:
        #with self.reply_wait_condition:
        got_it = self.reply_wait_condition.wait_for(
            lambda: new_reply_id in self.reply_message and self.reply_message[new_reply_id] is not None,
            timeout
        )
        if not got_it:
            # The timeout has expired and no message was received
            return -2, None
        reply_msg = self.reply_message.pop(new_reply_id, None)
        if reply_msg is None:
            # This should not occur due to the lambda check, but added as a safety net
            return -1, None
        error = reply_msg.get('error', 0)
        if error != 0:
            str_msg = reply_msg.get('message', None)
            raise OperationError(error, str_msg)
        return 0, reply_msg

    def wait_for_backends(self):
        logging.debug("Waiting for join kafka reply topic")
        self.__broker.wait_for_reply_available()

    def get(self, pv_url: str, timeout: float = None):
        """ Perform the get operation
            raise OperationTimeout when timeout has expired
        """
        protocol, pv_name = self.parse_pv_url(pv_url)
        if protocol.lower() != "pva" and protocol.lower() != "ca":
            raise ValueError("The protocol need to be one of 'pva'  'ca'")
        
        new_reply_id = str(uuid.uuid1())
        fetched = False
        result = None
        with self.reply_wait_condition:
            # clear the reply message for the requested pv
            self.reply_message[new_reply_id] = None
            # send message to k2eg
            self.__broker.send_get_command(
                pv_url,
                new_reply_id
            )
            while(not fetched):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout during get operation for {pv_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    fetched = True
        if result is not None and pv_name in result:
            return result[pv_name]
        else:
            return result
                
    def put(self, pv_url: str, value: any, timeout: float = None):
        """ Set the value for a single pv
        Args:
            pv_name   (str): is the name of the pv
            value     (str): is the new value
            protocol  (str): the protocol of the pv, the default is pva
            timeout (float): the timeout, in second or fraction
        Raises:
            ValueError: if some parameter are not valid
        
            return the error code and a message in case the error code is not 0
        """
        protocol, pv_name = self.parse_pv_url(pv_url)
        if protocol.lower() not in ("pva", "ca"):
            raise ValueError("The protocol need to be one of 'pva'  'ca'")

        # wait for consumer joined the topic
        fetched = False
        new_reply_id = str(uuid.uuid1())
        logging.info("Send and wait for message")
        with self.reply_wait_condition:
            # init reply slot
            self.reply_message[new_reply_id] = None
            # send message to k2eg
            self.__broker.send_put_command(
                pv_url,
                value,
                new_reply_id
            )
            while(not fetched):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout during start get operation for {pv_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    return result
    

    def monitor(self, pv_url: str, handler: Callable[[str, dict], None], timeout: float = None):  # noqa: E501
        """ Add a new monitor for pv if it is not already activated
        Parameters
                ----------
                pv_name : str
                    The name of the PV to monitor
                handler: function
                    The handler to be called when a message is received
        Rais:
                ----------
                True: the monitor has been activated
                False: otherwhise
        """
        fetched = False
        protocol, pv_name = self.parse_pv_url(pv_url)
        if protocol.lower() not in ("pva", "ca"):
            raise ValueError("The portocol need to be one of 'pva'  'ca'")
        new_reply_id = str(uuid.uuid1())
        with self.reply_wait_condition:
            # init reply slot
            self.reply_message[new_reply_id] = None
            if pv_name in self.__monitor_pv_handler:
                logging.info(
                    f"Monitor already activate for pv {pv_name}")
                return
            # send message to k2eg from activate (only for last topics) 
            # monitor(just in case it is not already activated)
            self.__broker.send_start_monitor_command(
                pv_url,
                self.__normalize_pv_name(pv_name),
                new_reply_id,
            )

            while(not fetched):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout during start monitor operation for {pv_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    # all is gone ok i can register the handler and subscribe
                    self.__monitor_pv_handler[pv_name] = handler
                    self.__broker.add_topic(self.__normalize_pv_name(pv_name))
                    return result
        
    def monitor_many(self, pv_uri_list: list[str], handler: Callable[[str, dict], None], timeout: float = None):  # noqa: E501
        """ Add a new monitor for pv if it is not already activated
        Parameters
                ----------
                pv_uri_list : list[str]
                    The name of the PV to monitor
                handler: function
                    The handler to be called when a message is received
        Rais:
                ----------
                True: the monitor has been activated
                False: otherwhise
        """
        fetched = False
        self._check_pv_list(pv_uri_list)
        new_reply_id = str(uuid.uuid1())
        with self.reply_wait_condition:
            filtered_list_pv_uri = []
            # init reply slot
            for pv_uri in pv_uri_list:
                protocol, pv_name = self.parse_pv_url(pv_uri)
                self.reply_message[new_reply_id] = None
                if pv_name in self.__monitor_pv_handler:
                    logging.info(
                        f"Monitor already activate for pv {pv_name}")
                    continue
                filtered_list_pv_uri.append(pv_uri)
            
            if len(filtered_list_pv_uri)==0:
                return
            # send message to k2eg from activate (only for last topics) 
            # monitor(just in case it is not already activated)
            self.__broker.send_start_monitor_command_many(
                filtered_list_pv_uri,
                new_reply_id,
            )

            while(not fetched):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout during start monitor operation for {pv_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    # all is gone ok i can register the handler and subscribe
                    for pv_uri in filtered_list_pv_uri:
                        protocol, pv_name = self.parse_pv_url(pv_uri)
                        self.__monitor_pv_handler[pv_name] = handler
                        self.__broker.add_topic(self.__normalize_pv_name(pv_name))
                    return result
    
    def stop_monitor(self, pv_name: str):  # noqa: E501
        """ Remove movitor for a specific pv
        Parameters
                ----------
                pv_name : str
                    The name of the PV to monitor
        """
        with self.reply_wait_condition:
            # all is gone ok i can register the handler and subscribe
            del self.__monitor_pv_handler[pv_name]
            self.__broker.remove_topic(self.__normalize_pv_name(pv_name))

    def snapshot(self,  pv_uri_list: list[str], handler: Callable[[str, dict], None])->str:
        """ Perform the snapshot creation
        return the id to be used to match the snapthot returned asynchronously in the hanlder
        """
        #check if all the pv are wellformed
        self._check_pv_list(pv_uri_list)
        new_reply_id = str(uuid.uuid1())
        with self.reply_wait_condition:
            # Set the snapshot handler and initialize the snapshot results vector
            self.reply_snapsthot_message[new_reply_id] = Snapshot(handler=handler)

            # send message to k2eg fto execute snapshot
            self.__broker.send_snapshot_command(
                pv_uri_list,
                new_reply_id,
            )
        return new_reply_id
    
    def snapshot_recurring(self,  properties: SnapshotProperties, handler: Callable[[str, Dict[str, Any]], None], timeout: float = None):
        """
        Create a new recurring snapshot for a list of process variables (PVs).

        This method initiates a recurring snapshot operation for the specified PVs.
        It registers a handler to be called asynchronously when snapshot data is available.
        The method blocks until the snapshot is created and an acknowledgment is received from the server,
        or until the specified timeout is reached.

        Args:
            snapshot_name (str): The name to assign to the recurring snapshot.
            pv_uri_list (list[str]): List of PV URIs to include in the snapshot.
            handler (Callable[[str, dict], None]): Callback function to handle snapshot results.
                The handler receives the snapshot ID and a dictionary containing the snapshot data.
            timeout (float, optional): Maximum time to wait for the server acknowledgment, in seconds.
                If None, waits indefinitely.

        Returns:
            str: "ok" if the snapshot is successfully created and acknowledged.

        Raises:
            ValueError: If any PV URI is not well-formed or uses an unsupported protocol.
            OperationTimeout: If the operation times out before receiving an acknowledgment.
            OperationError: If the server returns an error during snapshot creation.

        Example:
            def my_handler(snapshot_id, data):
                print(f"Snapshot {snapshot_id} data: {data}")

            dml_instance.snapshot_recurring(
                "my_snapshot",
                ["pva://my:pv1", "ca://my:pv2"],
                my_handler,
                timeout=5.0
            )
        """
        #check if all the pv are wellformed
        self._check_pv_list(properties.pv_uri_list)
        self._validate_snapshot_name(properties.snapshot_name)
        new_reply_id = str(uuid.uuid1())
        with self.reply_wait_condition:
            # init reply slot
            self.reply_message[new_reply_id] = None

            # create the snaphsot structure
            s = Snapshot(handler=handler)
            # fill pv_list with the name of the pv without pva:// or ca://
            s.pv_list = [
                self.parse_pv_url(pv_uri)[1] for pv_uri in properties.pv_uri_list
            ]
            s.properties = properties
            s.publishing_topic = None

            # send message to k2eg fto execute snapshot
            self.__broker.send_repeating_snapshot_command(
                properties,
                new_reply_id,
            )

            while(True):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout during the submition of snapshot {properties.snapshot_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    #at this point we need to start listening to the right topic
                    if "publishing_topic" in result:
                         # Set the snapshot handler and initialize the snapshot results vector
                        s.publishing_topic = result["publishing_topic"]
                        self.reply_recurring_snapsthot_message[s.publishing_topic] = s
                        self.__broker.add_topic(s.publishing_topic)
                        logging.info(
                            f"Recurring snapshot {properties.snapshot_name} listening on topic {s.publishing_topic}"
                        )
                        
                    return result

    def snapshost_trigger(self, snapshot_name: str, timeout: float = None):
        """ Trigger a new publishing of a specific snapshot
        """
        self._validate_snapshot_name(snapshot_name)
        new_reply_id = str(uuid.uuid1())
        with self.reply_wait_condition:
            # init reply slot
            self.reply_message[new_reply_id] = None

            # send message to k2eg fto execute snapshot
            self.__broker.send_repeating_snapshot_trigger_command(
                snapshot_name,
                new_reply_id,
            )

            while(True):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout during triggering the snapshot {snapshot_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    return result

    def snapshot_stop(self, snapshot_name: str, timeout: float = None):
        """ Stop the snapshot operation
        """
        self._validate_snapshot_name(snapshot_name)
        new_reply_id = str(uuid.uuid1())
        with self.reply_wait_condition:
            # init reply slot
            self.reply_message[new_reply_id] = None
            # the snapshot map key is the kafaka topic directly
            topic_key_for_snapshot_to_remove = None
            for topic, snapshot in self.reply_recurring_snapsthot_message.items():
                if snapshot.properties.snapshot_name == snapshot_name:
                    topic_key_for_snapshot_to_remove = topic
                    break
            if topic_key_for_snapshot_to_remove is not None:
                del self.reply_recurring_snapsthot_message[topic_key_for_snapshot_to_remove]
                self.__broker.remove_topic(topic_key_for_snapshot_to_remove)
            
            # send message to k2eg fto execute snapshot
            self.__broker.send_repeating_snapshot_stop_command(
                snapshot_name,
                new_reply_id,
            )

            while(True):
                op_res, result =  self.__wait_for_reply(new_reply_id, timeout)
                if op_res == -2:
                    # raise timeout exception
                    raise OperationTimeout(
                            f"Timeout stopping the snapshot {snapshot_name}"
                            )
                elif op_res == -1:
                    continue
                else:
                    return result

    def snapshot_sync(self,  pv_uri_list: list[str], timeout: float = 10.0)->list[dict[str, Any]]:
        """ Perform the snapshot operation
        return the snapshot value synchronously
        """
        snapshot_id = None
        received_snapshot = None
        #check if all the pv are wellformed
        def internal_snapshot_handler(id, snapshot_data):
            nonlocal snapshot_id
            nonlocal received_snapshot
            if snapshot_id == id:
                received_snapshot = snapshot_data
        snapshot_id = self.snapshot(pv_uri_list, internal_snapshot_handler)     
        # wait for received_snapshot isnot None of timeout expired
        
        while(received_snapshot is None):
            # wait some millisecondos on this thread
            sleep(0.3)
            if timeout is not None:
                timeout = timeout - 0.3
                if timeout <= 0:
                    raise OperationTimeout(
                        f"Timeout during snapshot operation for {pv_uri_list}"
                    )
        return {'error' : 0, "snapshot": received_snapshot}

    def close(self):
        # signal thread to terminate
        if self.__thread is not None:
            self.__consume_data = False
            self.__thread.join()
        if self.__broker is not None:
            self.__broker.close()



from concurrent.futures import ThreadPoolExecutor
import logging
from typing import Any, Dict
from uuid import uuid1
import k2eg
from k2eg.broker import SnapshotProperties
from k2eg.dml import Snapshot, _filter_pv_uri
import time
import pytest
from unittest import TestCase

k: k2eg.dml = None
TestCase.maxDiff = None

@pytest.fixture(scope="session", autouse=True)
def my_setup(request):
    print('Enable k2eg')
    global k
    logging.basicConfig(
            format="[%(asctime)s %(levelname)-8s] %(message)s",
            level=logging.DEBUG,
        )
    k = k2eg.dml('test', 'app-test', uuid1())
    def fin():
        global k
        if k is not None:
              k.close()
    request.addfinalizer(fin)

def test_uri_protocol_ioc_attribute():
    protocol, resource = _filter_pv_uri('pva://ioa:a.attribute')
    assert protocol == 'pva', "protocol should not be None"
    assert resource == 'ioa:a.attribute', "protocol should not be None"

def test_uri_protocol_ioc_multilevel_attribute():
    protocol, resource = _filter_pv_uri('pva://ioa:a.attribute.sub-attribute')
    assert protocol == 'pva', "protocol should not be None"
    assert resource == 'ioa:a.attribute.sub-attribute', "protocol should not be None"

def test_uri_protocol_ioc():
    protocol, resource = _filter_pv_uri('pva://ioa:a')
    assert protocol == 'pva', "protocol should not be None"
    assert resource == 'ioa:a', "protocol should not be None"

def test_uri_protocol_ioc_uppercase_and_dash():
    protocol, resource = _filter_pv_uri('ca://VGXX-L3B-1602-PLOG')
    assert protocol == 'ca', "protocol should not be None"
    assert resource == 'VGXX-L3B-1602-PLOG', "protocol should not be None"


def test_exception_on_get_without_protocol():
    with pytest.raises(ValueError, 
                    match=r".*url.*not.*formed"):
                    k.get('//bad:pv:name', timeout=0.5)


def test_exception_on_get_with_bad_pv_name():
    with pytest.raises(ValueError, 
                    match=r".*url.*not.*formed"):
                    k.get('pva://', timeout=0.5)

def test_exception_on_get_with_bad_protocol():
    with pytest.raises(ValueError, 
                    match=r".*url.*not.*formed"):
                    k.get('unkonwn://', timeout=0.5)

def test_k2eg_get():
    get_value = k.get('pva://channel:ramp:ramp')
    assert get_value is not None, "value should not be None"

def test_k2eg_get_timeout():
    with pytest.raises(k2eg.OperationTimeout, 
                match=r"Timeout.*"):
                k.get('pva://bad:pv:name', timeout=0.5)

def test_k2eg_get_bad_pv():
    with pytest.raises(k2eg.OperationError):
                    k.get('pva://bad:pv:name')

def test_k2eg_monitor():
    retry = 0
    received_message = None

    def monitor_handler(pv_name, new_value):
        nonlocal received_message
        received_message = new_value
    try:
        k.monitor('pva://channel:ramp:rampa', monitor_handler)
        while received_message is None and retry < 20:
            retry = retry+1
            time.sleep(2)
    finally:
        k.stop_monitor("channel:ramp:rampa")
    assert received_message is not None, "value should not be None"

def test_k2eg_monitor_wrong():
    retry = 0
    received_message = None

    def monitor_handler(pv_name, new_value):
        nonlocal received_message
        received_message = new_value
    try:
        k.monitor('pva://channel:ramp:rampc', monitor_handler)
        while received_message is None and retry < 5:
            retry = retry+1
            time.sleep(2)
    finally:
        k.stop_monitor("channel:ramp:rampc")
    assert received_message is  None, "value be None"

def test_k2eg_monitor_many():
    retry = 0
    received_message_a = False
    received_message_b = False
    def monitor_handler(pv_name, new_value):
        nonlocal received_message_a
        nonlocal received_message_b
        if pv_name=='channel:ramp:rampa':
            received_message_a = True
        if pv_name=='channel:ramp:rampb':
            received_message_b = True
    try:
        k.monitor_many(['pva://channel:ramp:rampa', 'pva://channel:ramp:rampb'], monitor_handler)
        while (received_message_a is False or received_message_b is False) and retry < 20:
            retry = retry+1
            time.sleep(2)
    finally:
        k.stop_monitor("channel:ramp:rampa")
        k.stop_monitor("channel:ramp:rampb")
    assert received_message_a is not False, "value should not be None"
    assert received_message_b is not False, "value should not be None"

def test_put():
    k.put("pva://variable:a", 0)
    k.put("pva://variable:b", 0)
    time.sleep(1)
    res_get = k.get("pva://variable:sum")
    assert res_get['value'] == 0, "value should not be 0"
    k.put("pva://variable:a", 2)
    k.put("pva://variable:b", 2)
    #give some time to ioc to update
    time.sleep(1)
    res_get = k.get("pva://variable:sum")
    assert res_get['value'] == 4, "value should not be 0"

def test_multi_threading_put():
    put_dic={
        "pva://variable:a": 0,
        "pva://variable:b": 0
    }
    with ThreadPoolExecutor(10) as executor:
        for key, value in put_dic.items():
            executor.submit(put, key, value)
    time.sleep(5)
    res_get = k.get("pva://variable:sum")
    assert res_get['value'] == 0, "value should not be 0"
    put_dic={
        "pva://variable:a": 2,
        "pva://variable:b": 2
    }
    with ThreadPoolExecutor(10) as executor:
        for key, value in put_dic.items():
            executor.submit(put, key, value)
    #give some time to ioc to update
    time.sleep(5)
    res_get = k.get("pva://variable:sum")
    assert res_get['value'] == 4, "value should not be 0"

def put(key, value):
    try:
        k.put(key, value, 10)
        print(f"Put {key} with value {value}")
    except Exception as e:
        print(f"An error occured: {e}")

def test_put_timeout():
    with pytest.raises(k2eg.OperationTimeout, 
                    match=r"Timeout.*"):
                    k.put("pva://bad:pv:name", 0, timeout=0.5)


def test_put_wrong_device_timeout():
    with pytest.raises(k2eg.OperationError):
                    k.put("pva://bad:pv:name", 0)

def test_snapshot_on_simple_fixed_pv():
    retry = 0
    snapshot_id = None
    received_snapshot = None
    def snapshot_handler(id, snapshot_data):
        nonlocal snapshot_id
        nonlocal received_snapshot
        if snapshot_id == id:
            received_snapshot = snapshot_data

    try:
        snapshot_id = k.snapshot(['pva://variable:a', 'pva://variable:b'], snapshot_handler)
        while (received_snapshot is None ) and retry < 3:
            retry = retry+1
            time.sleep(2)
        # received_snapshot shuld be a dict with the snapshot data
        assert isinstance(received_snapshot, list), "value should be a list"
        assert received_snapshot[0]['variable:a'] is not None, "value should not be None"
        assert received_snapshot[1]['variable:b'] is not None, "value should not be None"
    except Exception as e:
        assert False, f"An error occured: {e}"

def test_snapshot_on_simple_fixed_pv_sync():
    try:
        received_snapshot = k.snapshot_sync(['pva://variable:a', 'pva://variable:b'])
        # received_snapshot shuld be a dict with the snapshot data
        assert isinstance(received_snapshot, dict), "value should be a dict"
        assert "error" in received_snapshot, "error should be in the snapshot"
        assert received_snapshot['error'] == 0, "error should be 0"
        assert "snapshot" in received_snapshot, "snapshot should be in the snapshot"
        assert isinstance(received_snapshot['snapshot'], list), "value should be a list"
        assert received_snapshot['snapshot'][0]['variable:a'] is not None, "value should not be None"
        assert received_snapshot['snapshot'][1]['variable:b'] is not None, "value should not be None"
    except Exception as e:
        assert False, f"An error occured: {e}"
        
def test_recurring_snapshot():
    retry = 0
    snapshot_name = "snap_1"
    received_snapshot:list[Snapshot] = []
    def snapshot_handler(id, snapshot_data:Dict[str, Any]):
        nonlocal snapshot_name
        nonlocal received_snapshot
        if snapshot_name == id:
            received_snapshot.append(snapshot_data)
            
    try:
        result = k.snapshot_stop(snapshot_name)
        print(result)
        time.sleep(1)
        result = k.snapshot_recurring(
            SnapshotProperties(
                snapshot_name = snapshot_name,
                time_window = 1000,
                repeat_delay = 0,
                pv_uri_list = ['pva://variable:a', 'pva://variable:b'],
                triggered=False,
            ),
            handler=snapshot_handler,
            timeout=10,
        )
        print(result)
        while (len(received_snapshot) == 0 ) and retry < 5:
            retry = retry+1
            time.sleep(5)
        k.snapshot_stop(snapshot_name)
        time.sleep(1)
        # received_snapshot shuld be a dict with the snapshot data
        assert len(received_snapshot) > 0, "snapshot should not be None"
        assert received_snapshot[0]['timestamp'] > 0, "timestamp should be valid"
        assert received_snapshot[0]['iteration'] > 0, "interation should be valid"
        assert 'variable:a' in received_snapshot[0], "variable:a need be contained into the snapshot"
        assert 'variable:b' in received_snapshot[0], "variable:b need be contained into the snapshot"
    except Exception as e:
        assert False, f"An error occured: {e}"
    finally:
        k.snapshot_stop(snapshot_name)
        time.sleep(1)
        
def test_recurring_snapshot_triggered():
    snapshot_name = "snap_1"
    received_snapshot:list[Snapshot] = []
    def snapshot_handler(id, snapshot_data:Snapshot):
        nonlocal snapshot_name
        nonlocal received_snapshot
        if snapshot_name == id:
            received_snapshot.append(snapshot_data)
            
    try:
        result = k.snapshot_stop(snapshot_name)
        print(result)
        time.sleep(1)
        result = k.snapshot_recurring(
            SnapshotProperties(
                snapshot_name = snapshot_name,
                time_window = 1000,
                repeat_delay = 0,
                pv_uri_list = ['pva://variable:a', 'pva://variable:b'],
                triggered=True,
            ),
            handler=snapshot_handler,
            timeout=10,
        )
        print(result)
        time.sleep(2)
        # send 2 snapshots
        result = k.snapshost_trigger(snapshot_name)
        print(result)
        time.sleep(2)
        result = k.snapshost_trigger(snapshot_name)
        print(result)
        time.sleep(2)
        k.snapshot_stop(snapshot_name)
        time.sleep(1)
        assert len(received_snapshot) == 2 , "snapshot should only one"
        assert received_snapshot[0]['timestamp'] > 0, "timestamp should be valid"
        assert received_snapshot[0]['iteration'] > 0, "interation should be valid"
        assert 'variable:a' in received_snapshot[0], "variable:a need be contained into the snapshot"
        assert 'variable:b' in received_snapshot[0], "variable:b need be contained into the snapshot"
    except Exception as e:
        assert False, f"An error occured: {e}"
    finally:
        k.snapshot_stop("snap_1")
        time.sleep(1)